# -*- coding: utf-8 -*-
from enum import Enum


class Tags(Enum):
    DARTS = "darts"
    DATA = "data"
    DATAFRAMES = "dataframes"
    DYNAMIC = "dynamic"
    FORECASTING = "forecasting"
    MODELS = "models"
    PANDAS = "pandas"
    PREPROCESSING = "preprocessing"
    TIME_SERIES = "time_series"
    TRANSFORMS = "tranforms"
